from functools import partial as _partial


def partial(*args, **kwargs):
    """
    sanic now requires function handlers to have a __name__,
    so we wrap functools.partial to also assign a __name__,
    which functools.partial objects don't have
    """
    p = _partial(*args, **kwargs)
    p.__name__ = args[0].__name__
    return p



"""
The shared callback logic

expects to be called as part of a class with the properties:
* instance.should_whitelist
* instance.category
* instance.analyzer
* instance.white_lister
"""
import re

from qai.issues.meta import get_meta_value


def _is_valid(instance):
    assert instance.should_whitelist is not None
    assert instance.category is not None
    assert instance.analyzer is not None


def _has_callable(instance, method):
    return hasattr(instance, method) and callable(getattr(instance, method))


def _filter_0distance(segment, issues):
    """
    filter issues to drop 0-distance edits
    :param issues: issues to filter
    :return: filter issues
    """
    filtered_issues = []
    for issue in issues:
        if issue["suggestions"]:
            suggestions = []
            for suggestion in issue["suggestions"]:
                # shouldn't lowercase, as this would prevent capitalization fixes
                # e.g. this would prevent 1-distance edits
                # but without it, this doesn't work, for unexplained reasons
                if (segment[issue["from"]: issue["until"]]) == suggestion:
                    continue
                suggestions.append(suggestion)

            if suggestions:
                issue["suggestions"] = suggestions
                filtered_issues.append(issue)
        else:  # some services don't offer suggestions, so no need to check
            filtered_issues.append(issue)
    return filtered_issues


def _extract_quotes(segment):
    quotes = []
    for m in re.finditer(r'"((?!").)+"', segment):
        quotes.append({'start': m.start(), 'end': m.end()})
    return quotes


def _inside_quote(quotes: list, issues):
    if len(quotes):
        issues_outside = []
        for issue in issues:
            drop_issue = False
            for quote in quotes:
                if issue['from'] >= quote['start'] and issue['until'] <= quote['end']:
                    drop_issue = True
                    break
            if not drop_issue:
                issues_outside.append(issue)
        return issues_outside
    return issues


def qallback(instance, segment, metadata, language, content):
    """
    Helper libs won't memoize this for us, because
    "Arguments to memoized function must be hashable"
    so we need to build it ourselves... weakass Python can't hash class instances
    """
    _is_valid(instance)

    if language not in instance.supported_languages:
        print(
            f"Received language {language}, supported Languages: {instance.supported_languages}"
        )
        return []

    # first param is now dealt with by mediator, we don't have to look at it
    _, meta_value, sub_groups = get_meta_value(instance.category, metadata)

    if _has_callable(instance.analyzer, "analyze_with_full_meta"):
        issues = instance.analyzer.analyze_with_full_meta(segment, meta=metadata)
    elif _has_callable(instance.analyzer, "analyze_with_metadata"):
        issues = instance.analyzer.analyze_with_metadata(
            segment, language=language, meta_value=meta_value, sub_groups=sub_groups
        )
    elif _has_callable(instance.analyzer, 'analyze_with_full_payload'):
        # meta_value not implemented in this method, as it is removed from payload
        # TODO clean up meta_value code
        issues = instance.analyzer.analyze_with_full_payload(
            segment, language=language, sub_groups=sub_groups, meta=metadata, content=content
        )
    else:
        issues = instance.analyzer.analyze(segment)

    if instance.should_whitelist:
        # Should we have whitelisting on? This is a per-service setting;
        # e.g. Passive voice => no whitelisting
        # Gender bias => whitelisting
        issues = instance.white_lister(issues, meta_value, sub_groups)

    if len(issues):
        # filter issues to drop 0-distance edits
        if isinstance(issues[0], list):  # issues is List[List[dict]]
            filtered_issues = []
            for issues_list, sentence in zip(issues, segment):
                filtered_0distance_issues = _filter_0distance(sentence, issues_list)
                # drop issues inside quotes
                if instance.ignore_inside_quotes:
                    filtered_issues.append(_inside_quote(_extract_quotes(sentence), filtered_0distance_issues))
                else:
                    filtered_issues.append(filtered_0distance_issues)
        else:  # List[dict]
            # drop issues inside quotes
            if instance.ignore_inside_quotes:
                filtered_issues = _inside_quote(_extract_quotes(segment), _filter_0distance(segment, issues))
            else:
                filtered_issues = _filter_0distance(segment, issues)
        return filtered_issues
    else:
        return issues



from typing import List


class Segment:

    offset: int
    length: int
    segment_id: str
    content: str

    def __init__(self, offset: int, length: int, segment_id: str, content: str):
        self.offset = offset
        self.length = length
        self.segment_id = segment_id
        self.content = content

    def __str__(self) -> str:
        if self.content:
            return self.content
        return super().__str__()


class Paragraph:

    offset: int
    length: int
    content: str
    segments: List[Segment]

    def __init__(self, offset: int, length: int, content: str, segments: List[Segment]):
        self.offset = offset
        self.length = length
        self.content = content
        self.segments = segments

    def __str__(self) -> str:
        if self.content:
            return self.content
        return super().__str__()


class Document:

    content: str
    segments: List[Segment]
    paragraphs: List[Paragraph]

    def __init__(self, json_response):
        self.segments = []
        self.paragraphs = []

        if json_response:
            self.content = json_response['content']

            paragraph_segments = []
            offset = None
            length = 0
            pcontent = ''

            for segment in json_response['segments']:
                scontent = self.content[segment['offset']:segment['offset'] + segment['length']]
                seg = Segment(segment['offset'], segment['length'], segment['segmentId'], scontent)
                self.segments.append(seg)
                paragraph_segments.append(seg)

                pcontent += scontent
                length += segment['length']
                if offset is None:
                    offset = segment['offset']

                if scontent.endswith('\n'):
                    self.paragraphs.append(Paragraph(offset, length, pcontent, paragraph_segments))
                    paragraph_segments = []
                    offset = None
                    length = 0
                    pcontent = ''
        else:
            self.content = ''

    def __str__(self) -> str:
        if self.content:
            return self.content
        return super().__str__()



import traceback

from qai.qconnect.qallback import qallback


def parse(el):
    segment = el["content"]["segment"]
    metadata = el["chain"]["meta"]
    language = el["content"]["languageCode"].split("-")[0]
    content = el['content']
    return segment, metadata, language, content


def add_issues(el, issues):
    el["issues"] = issues
    return el


def mock_batch_output(els):
    # Mock no issue batch prediction output
    output_els = []
    for el in els:
        el = add_issues(el, [])  # issues = []
        output_els.append(el)
    return output_els


def add_issues_format_insensitive(instance, el, validator, debug=False, verbose=False):
    # We should pull some of this code out soon
    # Why write it then? because I don't trust the calling services
    # to do what Yakiv says will happen
    segment, metadata, language, content = parse(el)

    # try except in case of unforseen problems
    issues = []
    try:
        # envisioned scenario
        # here we would like the validation
        # if validator says "no, shall not pass"
        # -> we don't callback, just return issues = []
        if verbose:
            print(f"Segment: {segment}")

        if validator(segment):
            # if not validator.has_html(segment):
            issues = qallback(instance, segment, metadata, language, content)
    except:
        print(f"Error log for segment {segment}")
        print(traceback.format_exc())

        # Qallback already failed
        # In debug mode we want to see errors so we call it again
        if debug:
            issues = qallback(instance, segment, metadata, language, content)
    finally:
        el = add_issues(el, issues)

    return el


def add_issues_format_insensitive_batch(instance, els, debug=False, verbose=False):
    # Format 1-by-1 but pass batched
    # Dependand service should take care of metdata and language
    # Consistent issue format, otherwise empty response
    # Assumption: same meta and language

    input_els = []
    output_els = []

    for el in els:
        segment, metadata, language, content = parse(el)
        input_els.append(segment)

    try:
        if verbose:
            print(f"Segment: {input_els}")
        el_issues = qallback(instance, input_els, metadata, language, content)
    except:
        print("error log for batch of segments, mocking empty response")
        print(traceback.format_exc())
        # Qallback already failed
        # In debug mode we want to see errors so we call it again
        if debug:
            qallback(instance, input_els, metadata, language, content)
        return mock_batch_output(els)

    for el_issue, el in zip(el_issues, els):
        output_els.append(add_issues(el, el_issue))

    return output_els



import uuid
from typing import Any, Dict


def issue_to_v2(i: Dict[str, Any]) -> Dict[str, Any]:
    """
    Helper function to turn v1 issues into v2. Pretty similar formats
    Some keys were removed, visible became mandatory
    modifies the dict in place, redundant return
    """
    if "id" in i:
        del i["id"]
    if "issueId" in i:
        del i["issueId"]
    if "visible" not in i:
        i["visible"] = True
    return i


def make_issue(
    word: str,
    start: int,
    end: int,
    issue_type: str,
    score,
    simpleDescription: str,
    description="",
    suggestions=[],
    subCategory="",
    learnMore="",
    paragraph=False,
    header="",
    metaDict={},
    accept_all_changes=False,
    debug=False,
    segment=None,
    visible=True,
):
    """
    reference:
    https://qordoba.atlassian.net/wiki/spaces/HOME/pages/840368141/Content-AI

    Note on duplicate value issueId / id: the platform switched this key at some point
    But I don't trust being able to delete things, so I will just duplicate it, and one day during a refactor
    we can remove it :grimace:
    """
    _uuid = str(uuid.uuid4())
    if not description:
        description = f'"{word}" is correlated with {issue_type} language.'

    issue = {
        "issueId": _uuid,
        "id": _uuid,
        "issueType": issue_type,
        "from": start,
        "until": end,
        "score": score,
        "suggestions": suggestions,
        "description": description,
        "simpleDescription": simpleDescription,
        "meta": {"paragraph": paragraph},
        "visible": visible,
    }

    if subCategory:
        issue["meta"]["subCategory"] = subCategory

    if learnMore:
        issue["meta"]["learnMore"] = learnMore

    if header:
        issue["meta"]["header"] = header

    if accept_all_changes:
        issue["meta"]["acceptAllChanges"] = accept_all_changes

    if debug:
        # log warnings if meta_dict overwrites issue["meta"]
        common = [value for value in metaDict.keys() if value in issue["meta"]]
        for c in common:
            if metaDict[c] != issue["meta"][c]:
                print(
                    f"Meta_dict {c}:{metaDict[c]} is overwriting issue meta {c}:{issue['meta'][c]}"
                )

        if segment is not None:
            suggestion = suggestions[0] if len(suggestions) else ""
            issue["new_segment"] = segment[:start] + suggestion + segment[end:]

    issue["meta"].update(metaDict)

    return issue


def make_issue_v2(*args, **kwargs):
    return issue_to_v2(make_issue(*args, **kwargs))



import json
import os


def get_configs(config_file):
    if os.path.exists(config_file):
        return json.loads(open(config_file).read())
    else:
        raise FileNotFoundError(
            "Please path get_configs a full path to the config file"
        )



from distutils.version import LooseVersion

import requests


def get_pypi_versions():
    r = requests.get("https://pypi.org/pypi/qai/json")
    pypi_versions = list(r.json()["releases"].keys())
    pypi_versions.sort(key=LooseVersion)
    return pypi_versions



import re

from typing import List, Tuple


class Validator:
    html_pattern = re.compile("<.*?>")
    non_letter_pattern = re.compile("[^a-zA-Z'\-\s]")

    def __init__(self, ignore_html=True, ignore_token_fraction=0.5):
        self.ignore_html = ignore_html
        self.acceptable_fraction_of_ignorable_tokens = ignore_token_fraction

    def _has_html(self, segment: str) -> bool:
        return self.html_pattern.match(segment) != None

    def _has_non_letter(self, segment: str):
        return self.non_letter_pattern.match(segment) != None

    def _is_empty(self, segment: str):
        return len(segment.strip()) == 0

    def _is_unacceptable(self, segment: str):
        if self._is_empty(segment):
            print("segment is empty")
            return True
        if self.ignore_html and self._has_html(segment):
            print("segment has HTML")
            return True
        return False

    def _is_ignored_tokens(self, token: str) -> bool:
        return self._has_non_letter(token)

    def _has_too_many_ignore_tokens(self, segment: str) -> bool:
        tokens = segment.strip().split(" ")
        token_length = len(tokens)
        ignored_tokens = [t for t in tokens if self._is_ignored_tokens(t)]
        ignored_length = len(ignored_tokens)
        if token_length == 0:
            return False
        else:
            return (
                ignored_length / token_length
                > self.acceptable_fraction_of_ignorable_tokens
            )

    def __call__(self, segment: str):
        if self._is_unacceptable(segment):
            return False
        if self._has_too_many_ignore_tokens(segment):
            print(f"too many non-letter tokens in {segment}")
            return False
        return True



import pytest

from qai.qconnect.qallback import _filter_0distance


# for when we expect a test to fail
xfail = pytest.mark.xfail

segment_good = "He visits the store."
segment_needs_uc = "he visits the store."
start, end = 0, 9

issues = [{"from": start, "until": end, "suggestions": ["He visits"]}]
issues_with_one_0distance_and_one_good_suggestion = [
    {"from": start, "until": end, "suggestions": ["He visits", "They visit"]}
]


@xfail(raises=KeyError)
def test_issue_without_suggestions():
    """
    This is a known bug that we should fix
    If an issue doesn't have a suggestions key, we get an unhandled exception
    """
    bad_issue = [{"from": start, "until": end}]
    assert _filter_0distance(segment_good, bad_issue)


def test_it_filters_0_distance():
    assert (
        _filter_0distance(segment_good, issues) == []
    ), "it should filter a no-op suggestion"


def test_it_only_filters_0_distance_when_there_are_multiple_suggestions():
    assert _filter_0distance(
        segment_good, issues_with_one_0distance_and_one_good_suggestion
    ) == [{"from": 0, "until": 9, "suggestions": ["They visit"]}]


def test_it_doesnt_filter_capitalization_correction():
    assert (
        _filter_0distance(segment_needs_uc, issues) == issues
    ), "It shouldn't filter capitalization correction"




import json
import os
import sys

from attrdict import AttrDict


def in_venv(dirname):
    if "/venv/" in dirname:
        return True, "/venv"
    elif "/env/" in dirname:
        return True, "/env"
    elif "/virtualenv/" in dirname:
        return True, "/virtualenv"
    else:
        return False, ""


def exists_and_is_file(path):
    return os.path.exists(path) and os.path.isfile(path)


def get_strings_json(string_path=None):
    string_paths = [
        ["app", "strings.json"],
        ["strings.json"],
        ["resources", "strings.json"],
        ["strings", "strings.json"],
        ["conf", "strings.json"],
        ["config", "strings.json"],
    ]
    STRING_FILE = None
    if string_path is None:
        ROOT_DIR = os.path.dirname(sys.modules["__main__"].__file__)
        is_virtual, splitter = in_venv(ROOT_DIR)
        if is_virtual:
            ROOT_DIR = ROOT_DIR.split(splitter)[0]
        for sp in string_paths:
            STRING_FILE = os.path.join(ROOT_DIR, *sp)
            if exists_and_is_file(STRING_FILE):
                break
    else:
        STRING_FILE = string_path

    if STRING_FILE is None:
        raise FileNotFoundError("Could not find strings.json file")

    with open(STRING_FILE, "r", encoding="utf-8") as f:
        return AttrDict(json.load(f))



from typing import List

import en_core_web_sm
import pytest
from spacy.tokens import Span

from qai.issues.spacy_factor import SpacyFactor


nlp = en_core_web_sm.load()


@pytest.mark.parametrize(
    "span,suggestions,corrected",
    [
        (
            nlp("A majority of apps will not require any changes")[:3],
            ["most", "most of"],
            ["Most", "Most of"],
        ),
        (nlp("A certain fraction of")[:], ["a portion"], ["A portion"],),
        (nlp("A sufficient number of elephants.")[:3], ["enough"], ["Enough"]),
        (nlp("Hey, you suck")[2:], ["I suck"], ["I suck"]),
    ],
)
def test_case_starts_with_a_or_I(
    span: Span, suggestions: List[str], corrected: List[str]
):
    fixed_suggestions = SpacyFactor.match_case(span, suggestions)
    assert (
        fixed_suggestions == corrected
    ), f"Should title case if span starts with {span[0].text}"




import asyncio
from typing import Optional

import httpx

from .document import Document


class RestClient(object):

    def _instantiate_client(self):
        return httpx.Client()

    def _instantiate_async_client(self):
        return httpx.AsyncClient()

    def get(self, url):
        return self._instantiate_client().get(url)

    async def get_async(self, url):
        async with self._instantiate_async_client() as client:
            return await client.get(url)

    def run(self, coroutine):
        return asyncio.run(coroutine)


class DocumentRest(RestClient):

    def fetch_document(self, url) -> Optional[Document]:
        response = self.get(url)
        if response.status_code == 200:
            return Document(response.json())
        return None

    def fetch_document_async(self, url) -> Optional[Document]:
        response = self.run(self.get_async(url))
        if response.status_code == 200:
            return Document(response.json())
        return None



import re
import string
import sys

from bs4 import BeautifulSoup
from google.cloud import bigtable
from google.cloud.bigtable import row_filters


class BigTable(object):
    def __init__(self, project, instance_id, table_id):
        self.project = project
        self.instance_id = instance_id
        self.table_id = table_id
        self.table = None
        self._instantiate_table()

    def _instantiate_table(self):
        if not self.table:
            # Create a Cloud Bigtable client.
            client = bigtable.Client(project=self.project)
            # Connect to an existing Cloud Bigtable instance.
            instance = client.instance(instance_id=self.instance_id)
            # Open an existing table.
            self.table = instance.table(table_id=self.table_id)
        return self.table


class DocumentBigTable(BigTable):
    def __init__(self, project, instance_id, table_id, column_family_id, column_qualifier):
        super().__init__(project, instance_id, table_id)
        self.column_family_id = column_family_id
        self.column_qualifier = column_qualifier

    def fetch_document_content(self, document_id, organization_id):
        # Create a filter to only retrieve the most recent version of the cell for each column accross entire row.
        row_filter = row_filters.CellsColumnLimitFilter(1)

        row_key = str(document_id)[::-1] + '_' + str(organization_id)
        row = self._instantiate_table().read_row(row_key.encode('utf-8'), row_filter)

        if row:
            column_id = self.column_qualifier.encode('utf-8')
            return row.cells[self.column_family_id][column_id][0].value.decode('utf-8')
        else:
            return None

    def fetch_document_text(self, document_id, organization_id):
        content_html = self.fetch_document_content(document_id, organization_id)
        if content_html:
            # strip html code
            soup = BeautifulSoup(content_html, 'lxml')
            return soup.get_text(' ', True)
        else:
            return None


class ExtensionBigTable(BigTable):
    def __init__(self, project, instance_id, table_id, column_family_id, column_qualifier):
        super().__init__(project, instance_id, table_id)
        self.column_family_id = column_family_id
        self.column_qualifier = column_qualifier

    @staticmethod
    def _to_string(number, radix):
        """
        Returns a string representation of the first argument in the radix specified by the second argument
        :param number: to be converted to a string.
        :param radix: the radix to use in the string representation.
        :return:
        """
        if not 2 <= radix <= 36:
            radix = 10

        if radix == 10:
            return str(number)

        buf = ['' for i in range(0, 33)]
        char_pos = 32
        negative = (number < 0)

        if not negative:
            number = -number

        digits = string.digits + string.ascii_lowercase
        while number <= -radix:
            buf[char_pos] = digits[int(abs(number) % radix)]
            number = int(number / radix)
            char_pos -= 1
        buf[char_pos] = digits[int(-number)]

        if negative:
            char_pos -= 1
            buf[char_pos] = '-'

        return ''.join(buf[char_pos:33])

    @staticmethod
    def _complement(key: str, length: int):
        """
        pad the key with zeros to the specified length
        :param key: the key to complement
        :param length: the full length
        :return: padded string with zeros of the specified length
        """
        return key.rjust(length, '0')

    @staticmethod
    def _reverse(key: str):
        """
        reverse key characters, e.g. 12345 becomes 54321
        :param key: the key to reverse
        :return: reversed key
        """
        return key[::-1]

    @staticmethod
    def _long_key_formatter(key):
        """
        format a key to 14 characters
        the first character represent the number sign, 0 for negative, 1 for positive
        the second part is a 36 radix representation of the key
        then pad the concatenation of those two parts to 14 characters
        :param key: the key to format
        :return: formatted key
        """
        length = 14
        return '0' + ExtensionBigTable._complement(ExtensionBigTable._to_string(sys.maxsize + key + 1, 36),
                                                   length - 1) if key < 0 else '1' + ExtensionBigTable._complement(
            ExtensionBigTable._to_string(key, 36), length - 1)

    @staticmethod
    def _twisted_long_key_formatter(key):
        """
        reverse the result of `~qai.ExtensionBigTable._long_key_formatter`
        :param key: the key to format
        :return: formatted key
        """
        return ExtensionBigTable._reverse(ExtensionBigTable._long_key_formatter(key))

    @staticmethod
    def _max_length_string_key_formatter(key: str, length: int = 100):
        """
        pad the specified key to the specified length
        :param key: the key to pad
        :param length: the length to pad to
        :return: padded key to the specified length
        """
        return ExtensionBigTable._complement(key, length)

    @staticmethod
    def _generate_key(document_id, organization_id, workspace_id, persona_id):
        """
        generate key for bigtable extension table, format agreed with extensions back-end
        :param document_id: the document id
        :param organization_id: the organization id
        :param workspace_id: the workspace id
        :param persona_id: the persona id
        :return: formatted key
        """
        return '|'.join([
            ExtensionBigTable._twisted_long_key_formatter(organization_id),
            ExtensionBigTable._twisted_long_key_formatter(workspace_id),
            ExtensionBigTable._max_length_string_key_formatter(document_id),
            ExtensionBigTable._long_key_formatter(persona_id)
        ])

    def fetch_document_content(self, document_id, organization_id, workspace_id, persona_id):
        # Create a filter to only retrieve the most recent version of the cell for each column accross entire row.
        row_filter = row_filters.CellsColumnLimitFilter(1)

        row_key = ExtensionBigTable._generate_key(document_id, organization_id, workspace_id, persona_id)
        row = self._instantiate_table().read_row(row_key.encode('utf-8'), row_filter)

        if row:
            column_id = self.column_qualifier.encode('utf-8')
            return row.cells[self.column_family_id][column_id][0].value.decode('utf-8')
        else:
            return None

    def fetch_document_text(self, document_id, organization_id, workspace_id, persona_id):
        content = self.fetch_document_content(document_id, organization_id, workspace_id, persona_id)
        if content:
            # strip new lines
            return re.sub('\\n+', ' ', content)
        else:
            return None



from qai.qonfigs import get_configs


class QRemoteConnection(object):
    def __init__(self, analyzer, category="", white_lister=None, config_file=None):
        self.configs = get_configs(config_file)

        # Only use name passed in as fallback
        try:
            self.category = self.configs["SERVICE_NAME"]
            print("Using SERVICE_NAME from config.json as category")
        except KeyError:
            self.category = category

        # this will crash if configs doesn't have SUPPORTED_LANGUAGES
        # that is intentional
        supported_languages = self.configs["SUPPORTED_LANGUAGES"]

        if isinstance(supported_languages, list):
            self.supported_languages = supported_languages
        elif isinstance(supported_languages, str):
            self.supported_languages = [supported_languages]
        else:
            raise TypeError(f"Config file {config_file} must have SUPPORTED_LANGUAGES as string or list of strings.")

        #  print(self.supported_languages)
        for lang in self.supported_languages:
            if lang.find("-") != -1:
                raise ValueError(
                    f"Language {lang} contains a '-', which is almost certainly an error. Correct values look like 'en'."
                )

        # Dependency Injection
        self.analyzer = analyzer
        if not hasattr(analyzer, 'analyze') \
                and not hasattr(analyzer, 'analyze_with_full_meta') \
                and not hasattr(analyzer, 'analyze_with_metadata') \
                and not hasattr(analyzer, 'analyze_with_full_payload'):
            raise ValueError('Analyzer instance must have "analyze" method.')

        if white_lister is not None:
            if not callable(white_lister):
                e = f"{white_lister.__name__} is not callable."
                raise TypeError(e)
            if not self.category:
                raise ValueError("Must specify a category to whitelist")
            self.should_whitelist = True
            self.white_lister = white_lister
        else:
            self.should_whitelist = False
        # end DI

    def get_future(self):
        raise NotImplementedError

    def connect(self):
        raise NotImplementedError
